# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from pathlib import Path

FIXTURES = Path(__file__).resolve().parent / "data"
